// @bun
// src/wrapped-row.ts
class WrappedRow {
  row;
  constructor(row) {
    this.row = row;
  }
  toJSON() {
    return JSON.stringify(this.row);
  }
  unwrap() {
    return this.row;
  }
}

// src/prequel.ts
import Database from "bun:sqlite";

// src/error.ts
class SchemaError extends Error {
  constructor(message, table, column) {
    if (typeof table === "string" && typeof column === "string") {
      super(`SchemaError: ${message} ON ${table}.${column}`);
    } else if (typeof table === "string") {
      super(`SchemaError: ${message} ON ${table}`);
    } else {
      super(`SchemaError: ${message}`);
    }
  }
}

// src/columns.ts
function generateColumnDefinitionSQL(name, colDef) {
  const parts = [`"${name}"`, colDef.type];
  if (colDef.primary) {
    parts.push("PRIMARY KEY");
  }
  if (!colDef.primary && !colDef.nullable) {
    parts.push("NOT NULL");
  }
  if (colDef.unique) {
    parts.push("UNIQUE");
  }
  if (colDef.autoincrement) {
    parts.push("AUTOINCREMENT");
  }
  if (colDef.default !== undefined) {
    parts.push(`DEFAULT (${colDef.default})`);
  }
  if (colDef.references.length) {
    parts.push(`REFERENCES ${colDef.references}`);
    if (colDef.cascade) {
      parts.push("ON DELETE CASCADE");
    }
  }
  if (colDef.misc.length) {
    parts.push(colDef.misc);
  }
  return parts.join(" ");
}
function expandColumnShorthand(col) {
  if (typeof col === "string") {
    return {
      type: col,
      nullable: false,
      primary: false,
      misc: "",
      references: "",
      cascade: false,
      default: undefined,
      unique: false,
      autoincrement: false
    };
  }
  return {
    type: col.type ?? "TEXT",
    nullable: col.nullable ?? false,
    primary: col.primary ?? false,
    misc: col.misc ?? "",
    references: col.references ?? "",
    cascade: col.cascade ?? false,
    default: col.default ?? undefined,
    unique: col.unique ?? false,
    autoincrement: col.autoincrement ?? false
  };
}
function normalizeColumns(cols) {
  const keys = Object.keys(cols);
  const retval = {};
  for (const key of keys) {
    const val = cols[key];
    if (val === undefined) {
      throw new Error("Invariant: Undefined column descriptor");
    }
    retval[key] = expandColumnShorthand(val);
  }
  return retval;
}

// src/table.ts
class PrequelIDNotFoundError extends Error {
  constructor(tableName, id) {
    super(`ID ${id} not found in ${tableName}`);
  }
}

class PrequelEmptyResultsError extends Error {
  constructor() {
    super("No results found on a query which requires results");
  }
}
function shapeForQuery(obj) {
  const asInsertDataShape = {};
  const keys = Object.keys(obj);
  for (const key of keys) {
    const val = obj[key];
    if (typeof val !== "string" && typeof val !== "number" && val !== null) {
      throw new SchemaError(`Invalid value in query: ${key}: ${val} (type: ${typeof val})`);
    }
    const asQueryKey = `$${key}`;
    asInsertDataShape[asQueryKey] = val;
  }
  return asInsertDataShape;
}

class Table {
  _name;
  _db;
  _columnDefinitions;
  _columnNames;
  _createTableSQL;
  _primaryColumnName = "ROWID";
  _insertQuery;
  _updateQuery;
  _findQuery;
  _deleteQuery;
  _deleteRowQuery;
  _sizeQuery;
  _truncateQuery;
  query;
  constructor(db, name, cols) {
    if (db instanceof Prequel) {
      this._db = db.db;
    } else {
      this._db = db;
    }
    this._name = name;
    this._columnDefinitions = normalizeColumns(cols);
    this._columnNames = Object.keys(this._columnDefinitions);
    this.query = this._db.query.bind(this._db);
    const columnSQLParts = [];
    const updateColumnSQLParts = [];
    for (const columnName of this._columnNames) {
      const columnDefinition = this._columnDefinitions[columnName];
      if (columnDefinition.primary && this._primaryColumnName === "ROWID") {
        this._primaryColumnName = columnName;
      }
      if (!columnDefinition.primary) {
        updateColumnSQLParts.push(`"${columnName}" = $${columnName}`);
      }
      columnSQLParts.push(generateColumnDefinitionSQL(columnName, columnDefinition));
    }
    const columnSQL = columnSQLParts.join(", ");
    const updateColumnSQL = updateColumnSQLParts.join(", ");
    const allCols = this._columnNames.join(", ");
    const allColVars = this._columnNames.map((colName) => `$${colName}`).join(", ");
    this._createTableSQL = `CREATE TABLE IF NOT EXISTS "${name}" (${columnSQL});`;
    const insertSql = `INSERT INTO "${name}" (${allCols}) VALUES (${allColVars}) RETURNING *;`;
    const updateSql = `UPDATE "${name}" SET ${updateColumnSQL} WHERE "${this._primaryColumnName}" = $${this._primaryColumnName} RETURNING *;`;
    const getByIdSql = `SELECT * FROM "${name}" WHERE "${this._primaryColumnName}" = ? LIMIT 1;`;
    const delByIdSql = `DELETE FROM "${name}" WHERE "${this._primaryColumnName}" = ?;`;
    const deleteRowSql = `DELETE FROM "${name}" WHERE "${this._primaryColumnName}" = $${this._primaryColumnName};`;
    const truncateQuery = `DELETE FROM "${name}";`;
    this.db.exec(this._createTableSQL);
    this._insertQuery = this.db.query(insertSql);
    this._updateQuery = this.db.query(updateSql);
    this._findQuery = this.db.query(getByIdSql);
    this._deleteQuery = this.db.query(delByIdSql);
    this._sizeQuery = this.db.query(`SELECT count(*) as count FROM ${name}`);
    this._deleteRowQuery = this.db.query(deleteRowSql);
    this._truncateQuery = this.db.query(truncateQuery);
    if (db instanceof Prequel) {
      if (!db.hasTable(name)) {
        db.attachTable(this);
      }
    }
  }
  hasColumn(colName) {
    return this._columnNames.includes(colName);
  }
  reference(colName) {
    if (colName === undefined) {
      return `${this._name}(${this._primaryColumnName})`;
    }
    const asString = String(colName);
    if (!this.hasColumn(asString)) {
      throw new SchemaError("No such column available for reference", this._name, asString);
    }
    return `${this._name}(${asString})`;
  }
  insert(data) {
    const asInsertDataShape = shapeForQuery(data);
    return this._insertQuery.get(asInsertDataShape);
  }
  insertPartial(data) {
    const asInsertDataShape = shapeForQuery(data);
    const keys = Object.keys(data);
    const colNames = keys.join(", ");
    const varNames = keys.map((keyName) => `$${keyName}`);
    const query = this._db.prepare(`INSERT INTO ${this} (${colNames}) VALUES (${varNames}) RETURNING *;`);
    return query.get(asInsertDataShape);
  }
  insertWithout(rowWithKeysOmitted) {
    return this.insertPartial(rowWithKeysOmitted);
  }
  update(data) {
    if (this._primaryColumnName === "ROWID") {
      throw new Error("Cannot use `Table.update()` without setting a primary column");
    }
    const asInsertDataShape = shapeForQuery(data);
    return this._updateQuery.get(asInsertDataShape);
  }
  updateWhere(data, conditions) {
    const keys = Object.keys(data);
    const setParts = keys.map((key) => `"${key}" = $${key}`);
    const setClause = setParts.join(", ");
    const whereKeys = Object.keys(conditions);
    const whereParts = whereKeys.map((key) => `"${key}" = $${key}`);
    const whereClause = whereParts.join(" AND ");
    const query = this._db.prepare(`UPDATE ${this} SET ${setClause} WHERE ${whereClause} RETURNING *;`);
    const shapedData = shapeForQuery({
      ...data,
      ...conditions
    });
    return query.all(shapedData);
  }
  updateAll(data) {
    const keys = Object.keys(data);
    const setParts = keys.map((key) => `"${key}" = $${key}`);
    const setClause = setParts.join(", ");
    const query = this._db.prepare(`UPDATE ${this} SET ${setClause} RETURNING *;`);
    return query.all(shapeForQuery(data));
  }
  exists(id) {
    return this.findOneById(id) !== null;
  }
  existsWhere(conditions) {
    return this.findOneWhere(conditions) !== null;
  }
  countWhere(conditions) {
    const keys = Object.keys(conditions);
    const whereParts = keys.map((key) => `"${key}" = $${key}`);
    const whereClause = whereParts.join(" AND ");
    const query = this._db.prepare(`SELECT count(*) as count FROM ${this} WHERE ${whereClause};`);
    const shapedData = shapeForQuery(conditions);
    const countResult = query.get(shapedData) ?? { count: 0 };
    return countResult.count;
  }
  findOneById(id) {
    return this._findQuery.get(`${id}`);
  }
  findOneByIdOrFail(id) {
    const found = this.findOneById(id);
    if (!found) {
      throw new PrequelIDNotFoundError(this.name, id);
    }
    return found;
  }
  findAll() {
    return this._db.prepare(`SELECT * FROM ${this}`).all();
  }
  findOneWhere(conditions) {
    const keys = Object.keys(conditions);
    const whereParts = keys.map((key) => `"${key}" = $${key}`);
    const whereClause = whereParts.join(" AND ");
    const query = this._db.prepare(`SELECT * FROM ${this} WHERE ${whereClause};`);
    return query.get(shapeForQuery(conditions));
  }
  findOneWhereOrFail(conditions) {
    const found = this.findOneWhere(conditions);
    if (!found) {
      throw new PrequelEmptyResultsError;
    }
    return found;
  }
  findAllWhere(conditions) {
    const keys = Object.keys(conditions);
    const whereParts = keys.map((key) => `"${key}" = $${key}`);
    const whereClause = whereParts.join(" AND ");
    const query = this._db.prepare(`SELECT * FROM ${this} WHERE ${whereClause};`);
    return query.all(shapeForQuery(conditions));
  }
  delete(data) {
    if (this._primaryColumnName === "ROWID") {
      throw new Error("Cannot use `Table.delete()` without setting a primary column");
    }
    const asDeleteDataShape = shapeForQuery(data);
    this._deleteRowQuery.run(asDeleteDataShape);
  }
  deleteById = (id) => {
    if (this._primaryColumnName === "ROWID") {
      throw new Error("Cannot use `Table.deleteById()` without setting a primary column");
    }
    this._deleteQuery.run(`${id}`);
  };
  deleteWhere(conditions) {
    const keys = Object.keys(conditions);
    const whereParts = keys.map((key) => `"${key}" = $${key}`);
    const whereClause = whereParts.join(" AND ");
    const query = this._db.prepare(`DELETE FROM ${this} WHERE ${whereClause};`);
    query.run(shapeForQuery(conditions));
  }
  deleteAll() {
    this._truncateQuery.run();
  }
  size() {
    return this.count();
  }
  count() {
    const res = this._sizeQuery.get();
    if (res === null) {
      throw new Error("Failed to count rows");
    }
    return res.count;
  }
  save(wrapped) {
    const row = wrapped.unwrap();
    this.update(row);
  }
  toString() {
    return `"${this._name}"`;
  }
  get name() {
    return this._name;
  }
  get db() {
    return this._db;
  }
  get columns() {
    return this._columnNames;
  }
  get schema() {
    return this._createTableSQL;
  }
  get idColumnName() {
    return this._primaryColumnName;
  }
}

// src/prequel.ts
class PrequelKVStore {
  table;
  constructor(pq, tableName) {
    this.table = pq.table(tableName, {
      kv_id: {
        type: "INTEGER",
        primary: true,
        autoincrement: true
      },
      key: "TEXT",
      value: "TEXT"
    });
  }
  put(key, value) {
    const lookup = this.table.findOneWhere({
      key
    });
    if (lookup === null) {
      this.table.insertPartial({
        key,
        value: JSON.stringify(value)
      });
    } else {
      lookup.value = JSON.stringify(value);
      this.table.update(lookup);
    }
  }
  get(key) {
    const lookup = this.table.findOneWhere({ key });
    if (lookup === null) {
      return null;
    }
    return JSON.parse(lookup.value);
  }
  has(key) {
    return this.table.findOneWhere({
      key
    }) !== null;
  }
  keys() {
    return this.table.findAll().map((row) => row.key);
  }
  values() {
    return this.table.findAll().map((row) => JSON.parse(row.value));
  }
}

class Prequel {
  db;
  kv;
  tables = {};
  query;
  uncachedQuery;
  constructor(filename = ":memory:") {
    this.db = new Database(filename);
    this.query = this.db.query.bind(this.db);
    this.uncachedQuery = this.db.prepare.bind(this.db);
    this.db.exec("PRAGMA foreign_keys=ON");
    this.kv = new PrequelKVStore(this, "PrequelManagedKVStore");
  }
  enableWAL() {
    this.db.exec("PRAGMA journal_mode=WAL;");
  }
  close() {
    this.db.close();
  }
  hasTable(name) {
    return Object.keys(this.tables).includes(name);
  }
  table(name, cols) {
    if (this.hasTable(name)) {
      throw new SchemaError("Duplicate table name", name);
    }
    const newTable = new Table(this, name, cols);
    this.tables[name] = newTable;
    return newTable;
  }
  attachTable(table) {
    if (this.hasTable(table.name)) {
      throw new SchemaError("Duplicate table name", table.name);
    }
    this.tables[table.name] = table;
  }
  transaction(transactionFn) {
    try {
      this.db.exec("BEGIN TRANSACTION;");
      const result = transactionFn();
      this.db.exec("COMMIT;");
      return result;
    } catch (error) {
      this.db.exec("ROLLBACK;");
      throw error;
    }
  }
}
export {
  WrappedRow,
  Table,
  Prequel
};
